import time

import mlx.core as core
import mlx.nn as nn
import mlx.optimizers.optimizers as optimizers
import numpy as np
from sklearn.model_selection import train_test_split
from tabulate import tabulate


class NeuralNet:
    def __init__(
        self,
        module: nn.Module,
        criterion,
        optimizer=optimizers.SGD,
        lr=0.1,
        max_epochs=10,
        batch_size=128,
    ) -> None:
        self.module = module
        self.criterion = criterion
        self.optimizer = optimizer(learning_rate=lr)
        self.lr = lr
        self.batch_size = batch_size
        self.max_epochs = max_epochs

    def batch_iterate(self, batch_size, X, y):
        perm = core.array(np.random.permutation(y.size))
        for s in range(0, y.size, batch_size):
            ids = perm[s : s + batch_size]
            yield X[ids], y[ids]

    def loss_fn(self, model, X, y):
        return self.criterion(model(X), y, reduction="mean")

    def fit(self, raw_X, raw_y, **kwargs):
        self.module = self.module()
        self.module.train(mode=True)
        core.eval(self.module.parameters())
        loss_and_grad_fn = nn.value_and_grad(self.module, self.loss_fn)

        raw_X, X_test, raw_y, y_test = train_test_split(raw_X, raw_y)

        headers = [
            "epoch",
            "train_loss",
            "valid_acc",
            "valid_loss",
            "dur",
        ]
        metrics_list = []

        for epoch in range(1, self.max_epochs + 1):
            start_time = time.time()
            batch_losses = []
            valid_scores = []

            for X, y in self.batch_iterate(self.batch_size, raw_X, raw_y):
                loss, grads = loss_and_grad_fn(
                    self.module, core.array(X), core.array(y)
                )
                self.optimizer.update(model=self.module, gradients=grads)

                batch_losses.append(loss)

            avg_loss = np.mean(batch_losses)

            score = self.score(X_test, y_test)
            valid_scores.append(score)

            valid_loss, _ = loss_and_grad_fn(
                self.module, core.array(X_test), core.array(y_test)
            )

            total_time = time.time() - start_time

            metrics_list.append(
                [
                    f"{epoch}",
                    f"{avg_loss:.4f}",
                    f"{score:.4f}",
                    f"{float(valid_loss):.4f}",
                    f"{total_time:.4f}",
                ]
            )

            core.eval(self.module.parameters(), self.optimizer.state)

        print(tabulate(metrics_list, headers=headers))
        self.module.train(mode=False)

    def predict(self, X):
        return list(self.module(core.array(X)).argmax(axis=1))

    def predict_proba(self, X):
        pass



from typing import Callable

import mlx.optimizers.optimizers as optimizers
from mlx import nn
from sklearn.base import ClassifierMixin

from sklx.net import NeuralNet


class NeuralNetworkClassifier(NeuralNet, ClassifierMixin):
    module = None
    max_epochs = 10
    lr = 0.1
    batch_size = 10
    optimizer = optimizers.SGD

    def __init__(
        self,
        module: nn.Module,
        max_epochs: float,
        lr: float,
        criterion: Callable,
    ) -> None:
        self.module = module
        self.max_epochs = max_epochs
        self.lr = lr
        self.criterion = criterion
        self.optimizer = optimizers.SGD(learning_rate=lr)

    def fit(self, raw_X, raw_y, **kwargs):
        return super().fit(raw_X, raw_y, **kwargs)



import numpy as np
from mlx import nn
from mlx.nn import losses
from sklearn.datasets import make_classification

from sklx.classifier import NeuralNetworkClassifier


def test_neural_network_classifier():
    """
    This is just a simple test to make sure the basic usage works.
    """
    X, y = make_classification(1000, 20, n_informative=10, random_state=0)
    X = X.astype(np.float32)
    y = y.astype(np.int64)

    class MyModule(nn.Module):
        def __init__(self, num_units=10, nonlin=nn.ReLU()):
            super().__init__()
            self.layers = [
                nn.Linear(20, num_units),
                nonlin,
                nn.Dropout(0.5),
                nn.Linear(num_units, num_units),
                nn.Linear(num_units, 2),
                nn.LogSoftmax(),
            ]

        def __call__(self, X, **kwargs):
            for _, layer in enumerate(self.layers):
                X = layer(X)
            return X

    net = NeuralNetworkClassifier(
        MyModule, max_epochs=10, lr=0.1, criterion=losses.nll_loss
    )

    net.fit(X, y)
    net.predict_proba(X)
